/*
 * livepatch_bsc1193529
 *
 * Fix for CVE-2021-4028, bsc#1193529
 *
 *  Upstream commit:
 *  bc0bdc5afaa7 ("RDMA/cma: Do not change route.addr.src_addr.ss_family")
 *
 *  SLE12-SP3 commit:
 *  not affected
 *
 *  SLE12-SP4, SLE12-SP5, SLE15 and SLE15-SP1 commit:
 *  not affected
 *
 *  SLE15-SP2 commit:
 *  not affected
 *
 *  SLE15-SP3 commit:
 *  aac50f8d6ac81baf3a8a8194e76d1625a7d83fc7
 *
 *
 *  Copyright (c) 2021 SUSE
 *  Author: Nicolai Stange <nstange@suse.de>
 *
 *  Based on the original Linux kernel code. Other copyrights apply.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, see <http://www.gnu.org/licenses/>.
 */

#if !IS_MODULE(CONFIG_INFINIBAND)
#error "Live patch supports only CONFIG_INFINIBAND=m"
#endif

/* klp-ccp: from drivers/infiniband/core/cma.c */
#include <linux/in.h>
#include <linux/in6.h>
#include <linux/mutex.h>
#include <linux/random.h>
#include <linux/igmp.h>
#include <linux/xarray.h>
#include <linux/inetdevice.h>
#include <linux/slab.h>
#include <linux/module.h>
#include <net/route.h>
#include <net/net_namespace.h>

#include <net/ipv6.h>
#include <net/ip_fib.h>
#include <rdma/rdma_cm.h>

/* klp-ccp: from include/rdma/ib_addr.h */
static int (*klpe_rdma_addr_size)(const struct sockaddr *addr);

/* klp-ccp: from include/rdma/rdma_cm.h */
static void (*klpe_rdma_destroy_id)(struct rdma_cm_id *id);

static int (*klpe_rdma_bind_addr)(struct rdma_cm_id *id, struct sockaddr *addr);

int klpp_rdma_listen(struct rdma_cm_id *id, int backlog);

static __be64 (*klpe_rdma_get_service_id)(struct rdma_cm_id *id, struct sockaddr *addr);

/* klp-ccp: from drivers/infiniband/core/cma.c */
#include <rdma/rdma_netlink.h>
#include <rdma/ib.h>
#include <rdma/ib_cm.h>

/* klp-ccp: from include/rdma/ib_cm.h */
static struct ib_cm_id *(*klpe_ib_cm_insert_listen)(struct ib_device *device,
				     ib_cm_handler cm_handler,
				     __be64 service_id);

/* klp-ccp: from drivers/infiniband/core/cma.c */
#include <rdma/ib_sa.h>
#include <rdma/iw_cm.h>

/* klp-ccp: from include/rdma/iw_cm.h */
static struct iw_cm_id *(*klpe_iw_create_cm_id)(struct ib_device *device,
				 iw_cm_handler cm_handler, void *context);

static void (*klpe_iw_destroy_cm_id)(struct iw_cm_id *cm_id);

static int (*klpe_iw_cm_listen)(struct iw_cm_id *cm_id, int backlog);

/* klp-ccp: from drivers/infiniband/core/core_priv.h */
#include <linux/list.h>
#include <linux/spinlock.h>
#include <linux/cgroup_rdma.h>
#include <net/net_namespace.h>
#include <net/netns/generic.h>
#include <rdma/ib_verbs.h>
#include <rdma/opa_addr.h>
#include <rdma/ib_mad.h>
#include <rdma/restrack.h>
/* klp-ccp: from drivers/infiniband/core/mad_priv.h */
#include <linux/completion.h>
#include <linux/err.h>
#include <linux/workqueue.h>
#include <rdma/ib_mad.h>
#include <rdma/ib_smi.h>
#include <rdma/opa_smi.h>

/* klp-ccp: from drivers/infiniband/core/cma_priv.h */
enum rdma_cm_state {
	RDMA_CM_IDLE,
	RDMA_CM_ADDR_QUERY,
	RDMA_CM_ADDR_RESOLVED,
	RDMA_CM_ROUTE_QUERY,
	RDMA_CM_ROUTE_RESOLVED,
	RDMA_CM_CONNECT,
	RDMA_CM_DISCONNECT,
	RDMA_CM_ADDR_BOUND,
	RDMA_CM_LISTEN,
	RDMA_CM_DEVICE_REMOVAL,
	RDMA_CM_DESTROYING
};

struct rdma_id_private {
	struct rdma_cm_id	id;

	struct rdma_bind_list	*bind_list;
	struct hlist_node	node;
	struct list_head	list; /* listen_any_list or cma_device.list */
	struct list_head	listen_list; /* per device listens */
	struct cma_device	*cma_dev;
	struct list_head	mc_list;

	int			internal_id;
	enum rdma_cm_state	state;
	spinlock_t		lock;
	struct mutex		qp_mutex;

	struct completion	comp;
	refcount_t refcount;
	struct mutex		handler_mutex;

	int			backlog;
	int			timeout_ms;
	struct ib_sa_query	*query;
	int			query_id;
	union {
		struct ib_cm_id	*ib;
		struct iw_cm_id	*iw;
	} cm_id;

	u32			seq_num;
	u32			qkey;
	u32			qp_num;
	u32			options;
	u8			srq;
	u8			tos;
	u8			tos_set:1;
	u8                      timeout_set:1;
	u8			reuseaddr;
	u8			afonly;
	u8			timeout;
	enum ib_gid_type	gid_type;

	/*
	 * Internal to RDMA/core, don't use in the drivers
	 */
	struct rdma_restrack_entry     res;
	struct rdma_ucm_ece ece;
};

/* klp-ccp: from drivers/infiniband/core/cma.c */
static struct list_head (*klpe_dev_list);
static struct list_head (*klpe_listen_any_list);
static struct mutex (*klpe_lock);

struct cma_device {
	struct list_head	list;
	struct ib_device	*device;
	struct completion	comp;
	refcount_t refcount;
	struct list_head	id_list;
	enum ib_gid_type	*default_gid_type;
	u8			*default_roce_tos;
};

struct rdma_bind_list {
	enum rdma_ucm_port_space ps;
	struct hlist_head	owners;
	unsigned short		port;
};

static int (*klpe_cma_comp_exch)(struct rdma_id_private *id_priv,
			 enum rdma_cm_state comp, enum rdma_cm_state exch);

static inline struct sockaddr *cma_src_addr(struct rdma_id_private *id_priv)
{
	return (struct sockaddr *) &id_priv->id.route.addr.src_addr;
}

static inline bool cma_zero_addr(const struct sockaddr *addr)
{
	switch (addr->sa_family) {
	case AF_INET:
		return ipv4_is_zeronet(((struct sockaddr_in *)addr)->sin_addr.s_addr);
	case AF_INET6:
		return ipv6_addr_any(&((struct sockaddr_in6 *)addr)->sin6_addr);
	case AF_IB:
		return ib_addr_any(&((struct sockaddr_ib *)addr)->sib_addr);
	default:
		return false;
	}
}

static inline bool cma_loopback_addr(const struct sockaddr *addr)
{
	switch (addr->sa_family) {
	case AF_INET:
		return ipv4_is_loopback(
			((struct sockaddr_in *)addr)->sin_addr.s_addr);
	case AF_INET6:
		return ipv6_addr_loopback(
			&((struct sockaddr_in6 *)addr)->sin6_addr);
	case AF_IB:
		return ib_addr_loopback(
			&((struct sockaddr_ib *)addr)->sib_addr);
	default:
		return false;
	}
}

static inline bool cma_any_addr(const struct sockaddr *addr)
{
	return cma_zero_addr(addr) || cma_loopback_addr(addr);
}

static int (*klpe_cma_addr_cmp)(const struct sockaddr *src, const struct sockaddr *dst);

static int (*klpe_cma_ib_req_handler)(struct ib_cm_id *cm_id,
			      const struct ib_cm_event *ib_event);

static int (*klpe_iw_conn_req_handler)(struct iw_cm_id *cm_id,
			       struct iw_cm_event *iw_event);

static int klpr_cma_ib_listen(struct rdma_id_private *id_priv)
{
	struct sockaddr *addr;
	struct ib_cm_id	*id;
	__be64 svc_id;

	addr = cma_src_addr(id_priv);
	svc_id = (*klpe_rdma_get_service_id)(&id_priv->id, addr);
	id = (*klpe_ib_cm_insert_listen)(id_priv->id.device,
				 (*klpe_cma_ib_req_handler), svc_id);
	if (IS_ERR(id))
		return PTR_ERR(id);
	id_priv->cm_id.ib = id;

	return 0;
}

static int klpr_cma_iw_listen(struct rdma_id_private *id_priv, int backlog)
{
	int ret;
	struct iw_cm_id	*id;

	id = (*klpe_iw_create_cm_id)(id_priv->id.device,
			     (*klpe_iw_conn_req_handler),
			     id_priv);
	if (IS_ERR(id))
		return PTR_ERR(id);

	id->tos = id_priv->tos;
	id->tos_set = id_priv->tos_set;
	id_priv->cm_id.iw = id;

	memcpy(&id_priv->cm_id.iw->local_addr, cma_src_addr(id_priv),
	       (*klpe_rdma_addr_size)(cma_src_addr(id_priv)));

	ret = (*klpe_iw_cm_listen)(id_priv->cm_id.iw, backlog);

	if (ret) {
		(*klpe_iw_destroy_cm_id)(id_priv->cm_id.iw);
		id_priv->cm_id.iw = NULL;
	}

	return ret;
}

static int (*klpe_cma_listen_on_dev)(struct rdma_id_private *id_priv,
			     struct cma_device *cma_dev,
			     struct rdma_id_private **to_destroy);

static int klpr_cma_listen_on_all(struct rdma_id_private *id_priv)
{
	struct rdma_id_private *to_destroy;
	struct cma_device *cma_dev;
	int ret;

	mutex_lock(&(*klpe_lock));
	list_add_tail(&id_priv->list, &(*klpe_listen_any_list));
	list_for_each_entry(cma_dev, &(*klpe_dev_list), list) {
		ret = (*klpe_cma_listen_on_dev)(id_priv, cma_dev, &to_destroy);
		if (ret) {
			/* Prevent racing with cma_process_remove() */
			if (to_destroy)
				list_del_init(&to_destroy->list);
			goto err_listen;
		}
	}
	mutex_unlock(&(*klpe_lock));
	return 0;

err_listen:
	list_del(&id_priv->list);
	mutex_unlock(&(*klpe_lock));
	if (to_destroy)
		(*klpe_rdma_destroy_id)(&to_destroy->id);
	return ret;
}

static int klpr_cma_check_port(struct rdma_bind_list *bind_list,
			  struct rdma_id_private *id_priv, uint8_t reuseaddr)
{
	struct rdma_id_private *cur_id;
	struct sockaddr *addr, *cur_addr;

	lockdep_assert_held(&(*klpe_lock));

	addr = cma_src_addr(id_priv);
	hlist_for_each_entry(cur_id, &bind_list->owners, node) {
		if (id_priv == cur_id)
			continue;

		if (reuseaddr && cur_id->reuseaddr)
			continue;

		cur_addr = cma_src_addr(cur_id);
		if (id_priv->afonly && cur_id->afonly &&
		    (addr->sa_family != cur_addr->sa_family))
			continue;

		if (cma_any_addr(addr) || cma_any_addr(cur_addr))
			return -EADDRNOTAVAIL;

		if (!(*klpe_cma_addr_cmp)(addr, cur_addr))
			return -EADDRINUSE;
	}
	return 0;
}

int klpp_rdma_listen(struct rdma_cm_id *id, int backlog)
{
	struct rdma_id_private *id_priv =
		container_of(id, struct rdma_id_private, id);
	int ret;

	if (!(*klpe_cma_comp_exch)(id_priv, RDMA_CM_ADDR_BOUND, RDMA_CM_LISTEN)) {
		/*
		 * Fix CVE-2021-4028
		 *  +5 lines
		 */
		struct sockaddr_in any_in = {
			.sin_family = AF_INET,
			.sin_addr.s_addr = htonl(INADDR_ANY),
		};

		/* For a well behaved ULP state will be RDMA_CM_IDLE */
		/*
		 * Fix CVE-2021-4028
		 *  -2 lines, +1 line
		 */
		ret = (*klpe_rdma_bind_addr)(id, (struct sockaddr *)&any_in);
		if (ret)
			return ret;
		if (WARN_ON(!(*klpe_cma_comp_exch)(id_priv, RDMA_CM_ADDR_BOUND,
					   RDMA_CM_LISTEN)))
			return -EINVAL;
	}

	/*
	 * Once the ID reaches RDMA_CM_LISTEN it is not allowed to be reusable
	 * any more, and has to be unique in the bind list.
	 */
	if (id_priv->reuseaddr) {
		mutex_lock(&(*klpe_lock));
		ret = klpr_cma_check_port(id_priv->bind_list, id_priv, 0);
		if (!ret)
			id_priv->reuseaddr = 0;
		mutex_unlock(&(*klpe_lock));
		if (ret)
			goto err;
	}

	id_priv->backlog = backlog;
	if (id->device) {
		if (rdma_cap_ib_cm(id->device, 1)) {
			ret = klpr_cma_ib_listen(id_priv);
			if (ret)
				goto err;
		} else if (rdma_cap_iw_cm(id->device, 1)) {
			ret = klpr_cma_iw_listen(id_priv, backlog);
			if (ret)
				goto err;
		} else {
			ret = -ENOSYS;
			goto err;
		}
	} else {
		ret = klpr_cma_listen_on_all(id_priv);
		if (ret)
			goto err;
	}

	return 0;
err:
	id_priv->backlog = 0;
	/*
	 * All the failure paths that lead here will not allow the req_handler's
	 * to have run.
	 */
	(*klpe_cma_comp_exch)(id_priv, RDMA_CM_LISTEN, RDMA_CM_ADDR_BOUND);
	return ret;
}



#include <linux/kernel.h>
#include <linux/module.h>
#include "livepatch_bsc1193529.h"
#include "../kallsyms_relocs.h"

#define LIVEPATCHED_MODULE "rdma_cm"

static struct klp_kallsyms_reloc klp_funcs[] = {
	{ "ib_cm_insert_listen", (void *)&klpe_ib_cm_insert_listen, "ib_cm" },
	{ "rdma_addr_size", (void *)&klpe_rdma_addr_size, "ib_core" },
	{ "iw_cm_listen", (void *)&klpe_iw_cm_listen, "iw_cm" },
	{ "iw_create_cm_id", (void *)&klpe_iw_create_cm_id, "iw_cm" },
	{ "iw_destroy_cm_id", (void *)&klpe_iw_destroy_cm_id, "iw_cm" },
	{ "cma_addr_cmp", (void *)&klpe_cma_addr_cmp, "rdma_cm" },
	{ "cma_comp_exch", (void *)&klpe_cma_comp_exch, "rdma_cm" },
	{ "cma_ib_req_handler", (void *)&klpe_cma_ib_req_handler, "rdma_cm" },
	{ "cma_listen_on_dev", (void *)&klpe_cma_listen_on_dev, "rdma_cm" },
	{ "dev_list", (void *)&klpe_dev_list, "rdma_cm" },
	{ "iw_conn_req_handler", (void *)&klpe_iw_conn_req_handler, "rdma_cm" },
	{ "listen_any_list", (void *)&klpe_listen_any_list, "rdma_cm" },
	{ "lock", (void *)&klpe_lock, "rdma_cm" },
	{ "rdma_bind_addr", (void *)&klpe_rdma_bind_addr, "rdma_cm" },
	{ "rdma_destroy_id", (void *)&klpe_rdma_destroy_id, "rdma_cm" },
	{ "rdma_get_service_id", (void *)&klpe_rdma_get_service_id, "rdma_cm" },
};

static int livepatch_bsc1193529_module_notify(struct notifier_block *nb,
					      unsigned long action, void *data)
{
	struct module *mod = data;
	int ret;

	if (action != MODULE_STATE_COMING || strcmp(mod->name, LIVEPATCHED_MODULE))
		return 0;

	mutex_lock(&module_mutex);
	ret = __klp_resolve_kallsyms_relocs(klp_funcs, ARRAY_SIZE(klp_funcs));
	mutex_unlock(&module_mutex);
	WARN(ret, "livepatch: delayed kallsyms lookup failed. System is broken and can crash.\n");

	return ret;
}

static struct notifier_block livepatch_bsc1193529_module_nb = {
	.notifier_call = livepatch_bsc1193529_module_notify,
	.priority = INT_MIN+1,
};

int livepatch_bsc1193529_init(void)
{
	int ret;

	mutex_lock(&module_mutex);
	if (find_module(LIVEPATCHED_MODULE)) {
		ret = __klp_resolve_kallsyms_relocs(klp_funcs,
						    ARRAY_SIZE(klp_funcs));
		if (ret)
			goto out;
	}

	ret = register_module_notifier(&livepatch_bsc1193529_module_nb);
out:
	mutex_unlock(&module_mutex);
	return ret;
}

void livepatch_bsc1193529_cleanup(void)
{
	unregister_module_notifier(&livepatch_bsc1193529_module_nb);
}
